@@ -1,23 +1,25 @@
+# reshape2 dependency was originally abandoned because (1) we wanted to be in control
+#   of the R version dependency and (2) reshape2::dcast is not generic.
+#   reshape2 package is deprecated since December 2017, so we'll deprecate our
+#   redirection as well
 
-# melt is generic in reshape2 which is good (unlike dcast) but we still don't import reshape2 because reshape2's
-# dependency on R 3.1 could change in a future release of reshape2. Say it started to depend on R 3.3. Users of data.table
-# couldn't then install data.table in R 3.1 even if they only needed melt.data.table. The other reason is that
-# reshape2::dcast is not generic (see that method in fcast.R).
 melt <- function(data, ..., na.rm = FALSE, value.name = "value") {
   if (is.data.table(data)) {
     UseMethod("melt", data)
     # if data is not data.table and reshape2 is installed, this won't dispatch to reshape2's method;
     # CRAN package edarf and others fail without the else branch
+  # nocov start
   } else {
-    # nocov start
+    data_name = deparse(substitute(data))
     ns = tryCatch(getNamespace("reshape2"), error=function(e)
-         stop("The melt generic in data.table has been passed a ",class(data)[1L]," (not a data.table) but the reshape2 package is not installed to process this type. Please either install reshape2 and try again, or pass a data.table to melt instead."))
+      stop("The melt generic in data.table has been passed a ",class(data)[1L],", but data.table::melt currently only has a method for data.tables. Please confirm your input is a data.table, with setDT(", data_name, ") or as.data.table(", data_name, "). If you intend to use a method from reshape2, try installing that package first, but do note that reshape2 is deprecated and you should be migrating your code away from using it."))
+    warning("The melt generic in data.table has been passed a ", class(data)[1L], " and will attempt to redirect to the relevant reshape2 method; please note that reshape2 is deprecated, and this redirection is now deprecated as well. To continue using melt methods from reshape2 while both libraries are attached, e.g. melt.list, you can prepend the namespace like reshape2::melt(", data_name, "). In the next version, this warning will become an error.")
     ns$melt(data, ..., na.rm=na.rm, value.name=value.name)
-    # nocov end
   }
+  # nocov end
 }
 
-patterns <- function(..., cols=character(0L)) {
+patterns = function(..., cols=character(0L)) {
   # if ... has no names, names(list(...)) will be "";
   #   this assures they'll be NULL instead
   p = unlist(list(...), use.names = any(nzchar(names(...))))
@@ -26,7 +28,7 @@ patterns <- function(..., cols=character(0L)) {
   lapply(p, grep, cols)
 }
 
-melt.data.table <- function(data, id.vars, measure.vars, variable.name = "variable",
+melt.data.table = function(data, id.vars, measure.vars, variable.name = "variable",
        value.name = "value", ..., na.rm = FALSE, variable.factor = TRUE, value.factor = FALSE,
        verbose = getOption("datatable.verbose")) {
   if (!is.data.table(data)) stop("'data' must be a data.table")
@@ -34,14 +36,7 @@ melt.data.table <- function(data, id.vars, measure.vars, variable.name = "variab
   if (missing(measure.vars)) measure.vars = NULL
   measure.sub = substitute(measure.vars)
   if (is.call(measure.sub) && measure.sub[[1L]] == "patterns") {
-    measure.sub = as.list(measure.sub)[-1L]
-    idx = which(names(measure.sub) %in% "cols")
-    if (length(idx)) {
-      cols = eval(measure.sub[["cols"]], parent.frame())
-      measure.sub = measure.sub[-idx]
-    } else cols = names(data)
-    pats = lapply(measure.sub, eval, parent.frame())
-    measure.vars = patterns(pats, cols=cols)
+    measure.vars = do_patterns(measure.sub, names(data))
   }
   if (is.list(measure.vars) && length(measure.vars) > 1L) {
     meas.nm = names(measure.vars)
@@ -62,7 +57,7 @@ melt.data.table <- function(data, id.vars, measure.vars, variable.name = "variab
       value.name = meas.nm
     }
   }
-  ans <- .Call(Cfmelt, data, id.vars, measure.vars,
+  ans = .Call(Cfmelt, data, id.vars, measure.vars,
       as.logical(variable.factor), as.logical(value.factor),
       variable.name, value.name, as.logical(na.rm),
       as.logical(verbose))
