@@ -1,6 +1,21 @@
 # nocov start
 
-.onLoad <- function(libname, pkgname) {
+# used to raise message (write to STDERR but not raise warning) once per session only
+# in future this will be upgraded to warning, then error, until eventually removed after several years
+.pkg.store = new.env()
+.pkg.store$.unsafe.done = FALSE
+.unsafe.opt = function() {
+  if (.pkg.store$.unsafe.done) return(invisible())
+  val = getOption("datatable.nomatch")
+  if (is.null(val)) return(invisible())  # not set is ideal (it's no longer set in .onLoad)
+  if (identical(val, NA) || identical(val, NA_integer_)) return(invisible())  # set to default NA is ok for now; in future possible message/warning asking to remove
+  message("The option 'datatable.nomatch' is being used and is not set to the default NA. This option is still honored for now but will be deprecated in future. Please see NEWS for 1.12.4 for detailed information and motivation. To specify inner join, please specify `nomatch=NULL` explicitly in your calls rather than changing the default using this option.")
+  .pkg.store$.unsafe.done = TRUE
+}
+
+.Last.updated = vector("integer", 1L) # exported variable; number of rows updated by the last := or set(), #1885
+
+.onLoad = function(libname, pkgname) {
   # Runs when loaded but not attached to search() path; e.g., when a package just Imports (not Depends on) data.table
   if (!exists("test.data.table", .GlobalEnv, inherits=FALSE) &&    # check when installed package is loaded but skip when developing the package with cc()
       (dllV<-if(is.loaded("CdllVersion",PACKAGE="datatable")).Call(CdllVersion)else"before 1.12.0") != (RV<-packageVersion("data.table"))) {
@@ -18,7 +33,7 @@
   prefix = if (!missing(pkgname)) "data.table::" else ""  # R provides the arguments when it calls .onLoad, I don't in dev/test
   if (!length(grep("data.table", ss[[2L]], fixed = TRUE))) {
     ss = ss[c(1L, NA, 2L:length(ss))]
-    ss[[2L]] = parse(text=paste0("if (!identical(class(..1),'data.frame')) for (x in list(...)) { if (inherits(x,'data.table')) return(",prefix,"data.table(...)) }"))[[1]]
+    ss[[2L]] = parse(text=paste0("if (!identical(class(..1),'data.frame')) for (x in list(...)) { if (inherits(x,'data.table')) return(",prefix,"data.table(...)) }"))[[1L]]
     body(tt)=ss
     (unlockBinding)("cbind.data.frame",baseenv())
     assign("cbind.data.frame",tt,envir=asNamespace("base"),inherits=FALSE)
@@ -40,7 +55,6 @@
   # are relatively heavy functions where the overhead in getOption() would not be noticed.  It's only really [.data.table where getOption default bit.
   # Improvement to base::getOption() now submitted (100x; 5s down to 0.05s):  https://bugs.r-project.org/bugzilla/show_bug.cgi?id=17394
   opts = c("datatable.verbose"="FALSE",            # datatable.<argument name>
-       "datatable.nomatch"="NA_integer_",      # datatable.<argument name>
        "datatable.optimize"="Inf",             # datatable.<argument name>
        "datatable.print.nrows"="100L",         # datatable.<argument name>
        "datatable.print.topn"="5L",            # datatable.<argument name>
@@ -55,7 +69,7 @@
        "datatable.auto.index"="TRUE",          # DT[col=="val"] to auto add index so 2nd time faster
        "datatable.use.index"="TRUE",           # global switch to address #1422
        "datatable.prettyprint.char" = NULL,     # FR #1091
-       "datatable.old.unique.by.key" = "FALSE"  # TODO: change warnings in duplicated.R to error on or after May 2019 then remove a year after that.
+       "datatable.old.unique.by.key" = "FALSE"  # TODO: remove in May 2020
        )
   for (i in setdiff(names(opts),names(options()))) {
     eval(parse(text=paste0("options(",i,"=",opts[i],")")))
@@ -64,16 +78,6 @@
   if (!is.null(getOption("datatable.old.bywithoutby")))
     warning("Option 'datatable.old.bywithoutby' has been removed as warned for 2 years. It is now ignored. Please use by=.EACHI instead and stop using this option.")
 
-  # reshape2
-  # Tried this :
-  # if (!"package:reshape2" %in% search()) {
-  #   # temporary until reshape2 pull request to make generic is on CRAN ...
-  #   try(library(reshape2, pos="package:base", quietly=TRUE, warn.conflicts=FALSE), silent=TRUE)
-  # }
-  # which works. But then when melt in data.table is loaded, _that's_ what generates the mask message.
-  # There's also a NOTE: Package startup functions should not change the search path.
-  # Therefore, removed. Users will need to make sure reshape2 isn't loaded, or loaded behind data.table on search()
-
   # Test R behaviour that changed in v3.1 and is now depended on
   x = 1L:3L
   y = list(x)
@@ -102,16 +106,18 @@
   # R could feasibly in future not copy DF's vecsxp in this case. If that changes in R, we'd like to know via the warning
   # because tests will likely break too. The warning will quickly tell R-core and us why, so we can then update.
 
+  .Call(CinitLastUpdated, .Last.updated)  #1885
+
   invisible()
 }
 
-getRversion <- function(...) stop("Reminder to data.table developers: don't use getRversion() internally. Add a behaviour test to .onLoad instead.")
+getRversion = function(...) stop("Reminder to data.table developers: don't use getRversion() internally. Add a behaviour test to .onLoad instead.")
 # 1) using getRversion() wasted time when R3.0.3beta was released without the changes we expected in getRversion()>"3.0.2".
 # 2) R-devel and ourselves may wish to tinker with R-devel, turning on and off features in the same version number. So it's better if data.table doesn't hard code expectations into the version number.
-# 3) The discipline of adding a feaure test here helps fully understand the change.
+# 3) The discipline of adding a feature test here helps fully understand the change.
 # 4) Defining getRversion with a stop() here helps prevent new switches on getRversion() being added in future. Easily circumvented but the point is to issue the message above.
 
-.onUnload <- function(libpath) {
+.onUnload = function(libpath) {
   # fix for #474. the shared object name is different from package name
   # So 'detach' doesn't find datatable.so, as it looks by default for data.table.so
   library.dynam.unload("datatable", libpath)
