@@ -10,11 +10,16 @@ if (base::getRversion() < "3.5.0") {
   isTRUE  = function(x) is.logical(x) && length(x)==1L && !is.na(x) && x    # backport R's new implementation of isTRUE
   isFALSE = function(x) is.logical(x) && length(x)==1L && !is.na(x) && !x   # backport isFALSE that was added in R 3.5.0
 }
-isTRUEorNA    <- function(x) is.logical(x) && length(x)==1L && (is.na(x) || x)
-isTRUEorFALSE <- function(x) is.logical(x) && length(x)==1L && !is.na(x)
+isTRUEorNA    = function(x) is.logical(x) && length(x)==1L && (is.na(x) || x)
+isTRUEorFALSE = function(x) is.logical(x) && length(x)==1L && !is.na(x)
+allNA = function(x) .Call(C_allNAR, x)
+
+if (base::getRversion() < "3.2.0") {  # Apr 2015
+  isNamespaceLoaded = function(x) x %chin% loadedNamespaces()
+}
 
 # which.first
-which.first <- function(x)
+which.first = function(x)
 {
   if (!is.logical(x)) {
     stop("x not boolean")
@@ -23,7 +28,7 @@ which.first <- function(x)
 }
 
 # which.last
-which.last <- function(x)
+which.last = function(x)
 {
   if (!is.logical(x)) {
     stop("x not boolean")
@@ -31,47 +36,30 @@ which.last <- function(x)
   length(x) - match(TRUE, rev(x)) + 1L
 }
 
-# trim
-trim <- function(x) {
-  # Removes whitespace at the beginning and end of strings
-  # Assigning to x[] to retain the original dimensions, rownames and colnames
-  x[] = gsub(" +$", "", x)
-  x[] = gsub("^ +", "", x)
-  x
-}
-
-# take (I don't see it being used anywhere)
-take <- function(x, n=1L)
-{
-  # returns the head of head, without the last n observations
-  # convenient when inlining expressions
-  # NROW, for vectors, returns the vector length, but we cater for non-table like lists also here
-  # TO DO: allow taking along any dimension e.g. all but last column, rather than all but last row.
-  if (is.list(x) && !is.data.frame(x)  && !is.data.table(x)) l = length(x)
-  else l = NROW(x)
-  if (l < n) stop("Cannot take ",n," from ",l)
-  head(x, l-n)
-}
-# TODO: Implement take as UseMethod. Specific methods for each type.
-
-require_bit64 = function() {
-  # called in fread and print when they see integer64 columns are present
-  if (!requireNamespace("bit64",quietly=TRUE))
-  warning("Some columns are type 'integer64' but package bit64 is not installed. Those columns will print as strange looking floating point data. There is no need to reload the data. Simply install.packages('bit64') to obtain the integer64 print method and print the data again.")
+require_bit64_if_needed = function(DT) {
+  # called in fread and print.data.table
+  if (!isNamespaceLoaded("bit64") && any(sapply(DT,inherits,"integer64"))) {
+    # nocov start
+    # a test was attempted to cover the requireNamespace() by using unloadNamespace() first, but that fails when nanotime is loaded because nanotime also uses bit64
+    if (!requireNamespace("bit64",quietly=TRUE)) {
+      warning("Some columns are type 'integer64' but package bit64 is not installed. Those columns will print as strange looking floating point data. There is no need to reload the data. Simply install.packages('bit64') to obtain the integer64 print method and print the data again.")
+    }
+    # nocov end
+  }
 }
 
 # vapply for return value character(1)
-vapply_1c <- function (x, fun, ..., use.names = TRUE) {
+vapply_1c = function (x, fun, ..., use.names = TRUE) {
   vapply(X = x, FUN = fun, ..., FUN.VALUE = NA_character_, USE.NAMES = use.names)
 }
 
 # vapply for return value logical(1)
-vapply_1b <- function (x, fun, ..., use.names = TRUE) {
+vapply_1b = function (x, fun, ..., use.names = TRUE) {
   vapply(X = x, FUN = fun, ..., FUN.VALUE = NA, USE.NAMES = use.names)
 }
 
 # vapply for return value integer(1)
-vapply_1i <- function (x, fun, ..., use.names = TRUE) {
+vapply_1i = function (x, fun, ..., use.names = TRUE) {
   vapply(X = x, FUN = fun, ..., FUN.VALUE = NA_integer_, USE.NAMES = use.names)
 }
 
@@ -79,31 +67,33 @@ more = function(f) system(paste("more",f))    # nocov  (just a dev helper)
 
 # helper used to auto-name columns in data.table(x,y) as c("x","y"), CJ(x,y) and similar
 # naming of unnested matrices still handled by data.table()
-name_dots <- function(...) {
-  dot_sub <- as.list(substitute(list(...)))[-1L]
+name_dots = function(...) {
+  dot_sub = as.list(substitute(list(...)))[-1L]
   vnames = names(dot_sub)
   if (is.null(vnames)) {
-    vnames = rep.int("", length(dot_sub))
-    novname = rep.int(TRUE, length(dot_sub))
+    vnames = character(length(dot_sub))
   } else {
     vnames[is.na(vnames)] = ""
-    if (any(vnames==".SD")) stop("A column may not be called .SD. That has special meaning.")
-    novname = vnames==""
   }
-  for (i in which(novname)) {
-    if ((tmp <- deparse(dot_sub[[i]])[1L]) == make.names(tmp))
-      vnames[i] = tmp
+  notnamed = vnames==""
+  if (any(notnamed)) {
+    syms = sapply(dot_sub, is.symbol)  # save the deparse() in most cases of plain symbol
+    for (i in which(notnamed)) {
+      tmp = if (syms[i]) as.character(dot_sub[[i]]) else deparse(dot_sub[[i]])[1L]
+      if (tmp == make.names(tmp)) vnames[i]=tmp
+    }
   }
-  still_empty = vnames==""
-  if (any(still_empty)) vnames[still_empty] = paste0("V", which(still_empty))
-  list(vnames=vnames, novname=novname)
+  list(vnames=vnames, .named=!notnamed)
 }
 
 # convert a vector like c(1, 4, 3, 2) into a string like [1, 4, 3, 2]
 #   (common aggregation method for error messages)
-brackify = function(x) {
-  # arbitrary cutoff
-  if (length(x) > 10L) x = c(x[1:10], '...')
+brackify = function(x, quote=FALSE) {
+  # arbitrary
+  CUTOFF = 10L
+  # keep one more than needed to trigger dots if needed
+  if (quote && is.character(x)) x = paste0("'",head(x,CUTOFF+1L),"'")
+  if (length(x) > CUTOFF) x = c(x[1:CUTOFF], '...')
   sprintf('[%s]', paste(x, collapse = ', '))
 }
 
@@ -126,3 +116,9 @@ do_patterns = function(pat_sub, all_cols) {
 
   return(matched)
 }
+
+# nocov start #593 always return a data.table
+edit.data.table = function(name, ...) {
+  setDT(NextMethod('edit', name))[]
+}
+# nocov end
